/*
 *  sata_sis.c - Silicon Integrated Systems SATA
 *
 *  Maintained by:  Uwe Koziolek
 *  		    Please ALWAYS copy linux-ide@vger.kernel.org
 *		    on emails.
 *
 *  Copyright 2004 Uwe Koziolek
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  libata documentation is available via 'make {ps|pdf}docs',
 *  as Documentation/DocBook/libata.*
 *
 *  Hardware documentation available under NDA.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <scsi/scsi_host.h>
#include <linux/libata.h>

#define DRV_NAME	"sata_sis"
#define DRV_VERSION	"0.7-rtp"

enum {
	sis_180			= 0,
	sis_182			= 1,
	sis_1180		= 2,
	sis_1183		= 3,

	SIS_SCR_PCI_BAR		= 5,

	/* PCI configuration registers */
	SIS_GENCTL		= 0x54, /* IDE General Control register */
	SIS_SCR_BASE		= 0xc0, /* sata0 phy SCR registers */
	SIS180_SATA1_OFS	= 0x10, /* offset from sata0->sata1 phy regs */
	SIS182_SATA1_OFS	= 0x20, /* offset from sata0->sata1 phy regs */
	SIS_PMR			= 0x90, /* port mapping register */
	SIS_PMR_COMBINED	= 0x30,

	/* random bits */
	SIS_FLAG_CFGSCR		= (1 << 30), /* host flag: SCRs via PCI cfg */

	GENCTL_IOMAPPED_SCR	= (1 << 26), /* if set, SCRs are in IO space */
};

static struct {
	const char *name;
} sata_chip_name[] = {
	{ "SiS180/SiS181"   },
	{ "SiS182"   },
	{ "SiS1180"         },
	{ "SiS1182/SiS1183" }
};

struct sis_port_map
{
	u8 port2_start;
	u8 port2_scr_offset;
	u8 combined_mode;
	u8 slave;
};

static struct sis_port_map sis180_port_map = {
	.port2_start = 0,
	.port2_scr_offset = SIS180_SATA1_OFS,
	.combined_mode = 1,
	.slave = ATA_FLAG_SLAVE_POSS,
};

static struct sis_port_map sis180_sata_port_map = {
	.port2_start = 64,
	.port2_scr_offset = SIS180_SATA1_OFS,
	.combined_mode = 0,
};

static struct sis_port_map sis182_port_map = {
	.port2_start = 0x20,
	.port2_scr_offset = SIS182_SATA1_OFS,
	.combined_mode = 1,
};

static struct sis_port_map sis1180_port_map = {
	.port2_start = 0,
	.port2_scr_offset = SIS182_SATA1_OFS,
	.combined_mode = 0,
};

static struct sis_port_map sis1183_968_port_map = {
	.port2_start = 0x10,
	.port2_scr_offset = SIS182_SATA1_OFS,
	.combined_mode = 0,
};

static struct sis_port_map *port_map_table[] = {
	[sis_180]  = &sis180_port_map,
	[sis_182]  = &sis182_port_map,
	[sis_1180] = &sis1180_port_map,
	[sis_1183] = &sis182_port_map
};

static int sis_init_one (struct pci_dev *pdev, const struct pci_device_id *ent);
static u32 sis_scr_read (struct ata_port *ap, unsigned int sc_reg);
static void sis_scr_write (struct ata_port *ap, unsigned int sc_reg, u32 val);
static void sis_host_stop (struct ata_host_set *host_set);

static const struct pci_device_id sis_pci_tbl[] = {
	{ PCI_VENDOR_ID_SI, 0x180, PCI_ANY_ID, PCI_ANY_ID, 0, 0, sis_180 },
	{ PCI_VENDOR_ID_SI, 0x181, PCI_ANY_ID, PCI_ANY_ID, 0, 0, sis_180 },
	{ PCI_VENDOR_ID_SI, 0x182, PCI_ANY_ID, PCI_ANY_ID, 0, 0, sis_182 },
	{ PCI_VENDOR_ID_SI, 0x1180, PCI_ANY_ID, PCI_ANY_ID, 0, 0, sis_1180 },
	{ PCI_VENDOR_ID_SI, 0x1182, PCI_ANY_ID, PCI_ANY_ID, 0, 0, sis_1183 },
	{ PCI_VENDOR_ID_SI, 0x1183, PCI_ANY_ID, PCI_ANY_ID, 0, 0, sis_1183 },
	{ }	/* terminate list */
};


static struct pci_driver sis_pci_driver = {
	.name			= DRV_NAME,
	.id_table		= sis_pci_tbl,
	.probe			= sis_init_one,
	.remove			= ata_pci_remove_one,
};

static struct scsi_host_template sis_sht = {
	.module			= THIS_MODULE,
	.name			= DRV_NAME,
	.ioctl			= ata_scsi_ioctl,
	.queuecommand		= ata_scsi_queuecmd,
	.can_queue		= ATA_DEF_QUEUE,
	.this_id		= ATA_SHT_THIS_ID,
	.sg_tablesize		= ATA_MAX_PRD,
	.cmd_per_lun		= ATA_SHT_CMD_PER_LUN,
	.emulated		= ATA_SHT_EMULATED,
	.use_clustering		= ATA_SHT_USE_CLUSTERING,
	.proc_name		= DRV_NAME,
	.dma_boundary		= ATA_DMA_BOUNDARY,
	.slave_configure	= ata_scsi_slave_config,
	.bios_param		= ata_std_bios_param,
};

static const struct ata_port_operations sis_ops = {
	.port_disable		= ata_port_disable,
	.tf_load		= ata_tf_load,
	.tf_read		= ata_tf_read,
	.check_status		= ata_check_status,
	.exec_command		= ata_exec_command,
	.dev_select		= ata_std_dev_select,
	.phy_reset		= sata_phy_reset,
	.bmdma_setup            = ata_bmdma_setup,
	.bmdma_start            = ata_bmdma_start,
	.bmdma_stop		= ata_bmdma_stop,
	.bmdma_status		= ata_bmdma_status,
	.qc_prep		= ata_qc_prep,
	.qc_issue		= ata_qc_issue_prot,
	.eng_timeout		= ata_eng_timeout,
	.irq_handler		= ata_interrupt,
	.irq_clear		= ata_bmdma_irq_clear,
	.scr_read		= sis_scr_read,
	.scr_write		= sis_scr_write,
	.port_start		= ata_port_start,
	.port_stop		= ata_port_stop,
	.host_stop		= sis_host_stop,
};

static struct ata_port_info sis_port_info = {
	.sht		= &sis_sht,
	.host_flags	= ATA_FLAG_SATA | ATA_FLAG_SATA_RESET |
			  ATA_FLAG_NO_LEGACY,
	.pio_mask	= 0x1f,
	.mwdma_mask	= 0x7,
	.udma_mask	= 0x7f,
	.port_ops	= &sis_ops,
};


MODULE_AUTHOR("Uwe Koziolek");
MODULE_DESCRIPTION("low-level driver for Silicon Integratad Systems SATA controller");
MODULE_LICENSE("GPL");
MODULE_DEVICE_TABLE(pci, sis_pci_tbl);
MODULE_VERSION(DRV_VERSION);

static inline unsigned int get_scr_cfg_addr(struct ata_port *ap, unsigned int sc_reg)
{
	struct sis_port_map *port_map = ap->host_set->private_data;
	unsigned int addr = SIS_SCR_BASE + (4 * sc_reg);

	if (ap->port_no)
		addr += port_map->port2_scr_offset;

	return addr;
}

static u32 sis_scr_cfg_read (struct ata_port *ap, unsigned int sc_reg)
{
	struct pci_dev *pdev = to_pci_dev(ap->host_set->dev);
	struct sis_port_map *port_map = ap->host_set->private_data;
	unsigned int cfg_addr = get_scr_cfg_addr(ap, sc_reg);
	u32 val, val2 = 0;

	if (sc_reg == SCR_ERROR) /* doesn't exist in PCI cfg space */
		return 0xffffffff;

	pci_read_config_dword(pdev, cfg_addr, &val);
	if (port_map->combined_mode)
		pci_read_config_dword(pdev, cfg_addr+0x10, &val2);

	return (val|val2) & 0xfffffffb; /* avoid problems with powerdowned ports */
}

static void sis_scr_cfg_write (struct ata_port *ap, unsigned int scr, u32 val)
{
	struct pci_dev *pdev = to_pci_dev(ap->host_set->dev);
	struct sis_port_map *port_map = ap->host_set->private_data;
	unsigned int cfg_addr = get_scr_cfg_addr(ap, scr);

	if (scr == SCR_ERROR) /* doesn't exist in PCI cfg space */
		return;

	pci_write_config_dword(pdev, cfg_addr, val);

	if (port_map->combined_mode)
		pci_write_config_dword(pdev, cfg_addr+0x10, val);
}

static u32 sis_scr_read (struct ata_port *ap, unsigned int sc_reg)
{
	struct sis_port_map *port_map = ap->host_set->private_data;
	u32 val, val2 = 0;

	if (sc_reg > SCR_CONTROL)
		return 0xffffffffU;

	if (ap->flags & SIS_FLAG_CFGSCR)
		return sis_scr_cfg_read(ap, sc_reg);

	val = inl(ap->ioaddr.scr_addr + (sc_reg * 4));

	if (port_map->combined_mode)
		val2 = inl(ap->ioaddr.scr_addr + (sc_reg * 4) + 0x10);

	return (val | val2) & 0xfffffffb;
}

static void sis_scr_write (struct ata_port *ap, unsigned int sc_reg, u32 val)
{
	struct sis_port_map *port_map = ap->host_set->private_data;

	if (sc_reg > SCR_CONTROL)
		return;

	if (ap->flags & SIS_FLAG_CFGSCR)
		sis_scr_cfg_write(ap, sc_reg, val);
	else {
		outl(val, ap->ioaddr.scr_addr + (sc_reg * 4));
		if (port_map->combined_mode)
			outl(val, ap->ioaddr.scr_addr + (sc_reg * 4)+0x10);
	}
}

static void sis_init_port_map(struct pci_dev *pdev, const struct pci_device_id *ent, struct sis_port_map *port_map)
{
	u8 pmr;
	u32 val;
	char *bridge;

	memcpy(port_map,port_map_table[ent->driver_data],sizeof(*port_map));
	bridge = "";

	/* Detect if SiS180/181 is in combined mode or not */
	if (ent->driver_data == sis_180) {
		pci_read_config_byte(pdev, SIS_PMR, &pmr);
		if ((pmr & SIS_PMR_COMBINED) == 0) {
			memcpy(port_map, &sis180_sata_port_map, sizeof(*port_map));
		}
	}

	if (ent->driver_data == sis_182) {
		bridge="SiS965L";
		pci_read_config_dword ( pdev, 0x6C, &val);
		if (val & (1L << 31)) {
			bridge="SiS965";
			port_map->slave = ATA_FLAG_SLAVE_POSS;
		}
	}

	/* SiS 1180/1182 behave differently if we're on 966 or on 966L
	 *  read OFFSET 64h, bit 28 to decide  966 or 966L.
	 *  1: 966L port0 = c0h, port1 = e0h
	 *  0: 966  port0 = c0h, port1 = d0h
	 **/
	if (ent->driver_data >= sis_1180) {
		bridge="SiS966L";
		pci_read_config_dword(pdev, 0x64, &val);
		if (!(val & 0x10000000)) {
			bridge="SiS966";
			if (ent->driver_data == sis_1180)
				port_map->port2_scr_offset = SIS180_SATA1_OFS;
			port_map->slave = ATA_FLAG_SLAVE_POSS;
		}
	}

	/* SiS 1183 behaves differently on 966 and 968 chips.
	 * In 968, it only gets two SATA ports, one per channel.
	 * But in 966 chips, 1183 gets 4 SATA ports, and 2 for 966L
	 * 1183 SCRs offset are differernt between 966 and 968 chips
	 **/
	/* Note from RTP: I assume that 1182 on SiS968 behaves like 1183
	 * even if getting a board with such configuration is unlikely
	 **/
	if (ent->driver_data == sis_1183) {
	 	if (!pci_get_device(PCI_VENDOR_ID_SI, 0x0966, NULL))  {
			bridge="SiS968";
			memcpy(port_map, &sis1183_968_port_map, sizeof(*port_map));
		}
	}

	dev_printk(KERN_INFO, &pdev->dev, "Detected %s chipset in %s mode\n",
			   sata_chip_name[ent->driver_data].name,
			   port_map->combined_mode ? "combined" : "SATA");

	if (strcmp(bridge,""))
		dev_printk(KERN_INFO, &pdev->dev, "Detected a %s bridge\n",bridge);

}
static int sis_init_one (struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct sis_port_map *port_map;
	static int printed_version;
	struct ata_probe_ent *probe_ent = NULL;
	int rc;
	u32 genctl;
	struct ata_port_info *ppi = &sis_port_info;
	int pci_dev_busy = 0;
	u8 pci_class_prog;
	u16 pci_class;

	if (!printed_version++)
		dev_printk(KERN_INFO, &pdev->dev, "version " DRV_VERSION "\n");

	port_map = kzalloc(sizeof(*port_map),  GFP_KERNEL);
	if (!port_map)
		return -ENOMEM;

	/* check if the device's programming interface 09h,
	 * is compatibility mode
	 **/
	pci_read_config_word(pdev, PCI_CLASS_DEVICE, &pci_class);
	if (pci_class == PCI_CLASS_STORAGE_IDE) {
		pci_read_config_byte(pdev, PCI_CLASS_PROG, &pci_class_prog);
		if ((pci_class_prog & 0x5) == 0) {
			dev_printk(KERN_INFO, &pdev->dev,
				   "%s IDE device is in compatibiltiy mode.\n"
				   "I'll ignore it\n"
				   "Try to change BIOS setting with native mode.\n",
				   sata_chip_name[ent->driver_data].name);
			rc = -ENODEV;
			goto err_mem;
		}
	}

	rc = pci_enable_device(pdev);
	if (rc) {
		goto err_mem;
	}

	rc = pci_request_regions(pdev, DRV_NAME);
	if (rc) {
		pci_dev_busy = 1;
		goto err_out;
	}

	rc = pci_set_dma_mask(pdev, ATA_DMA_MASK);
	if (rc)
		goto err_out_regions;
	rc = pci_set_consistent_dma_mask(pdev, ATA_DMA_MASK);
	if (rc)
		goto err_out_regions;

	/* check and see if the SCRs are in IO space or PCI cfg space */
	pci_read_config_dword(pdev, SIS_GENCTL, &genctl);
	if ((genctl & GENCTL_IOMAPPED_SCR) == 0)
		ppi->host_flags |= SIS_FLAG_CFGSCR;

	/* if hardware thinks SCRs are in IO space, but there are
	 * no IO resources assigned, change to PCI cfg space.
	 */
	if ((!(ppi->host_flags & SIS_FLAG_CFGSCR)) &&
	    ((pci_resource_start(pdev, SIS_SCR_PCI_BAR) == 0) ||
	     (pci_resource_len(pdev, SIS_SCR_PCI_BAR) < 128))) {
		genctl &= ~GENCTL_IOMAPPED_SCR;
		pci_write_config_dword(pdev, SIS_GENCTL, genctl);
		ppi->host_flags |= SIS_FLAG_CFGSCR;
	}

	sis_init_port_map(pdev, ent, port_map);
	ppi->host_flags |= port_map->slave;

	probe_ent = ata_pci_init_native_mode(pdev, &ppi, ATA_PORT_PRIMARY | ATA_PORT_SECONDARY);
	if (!probe_ent) {
		rc = -ENOMEM;
		goto err_out_regions;
	}

	probe_ent->private_data = port_map;
	if (!(probe_ent->host_flags & SIS_FLAG_CFGSCR)) {
		probe_ent->port[0].scr_addr =
			pci_resource_start(pdev, SIS_SCR_PCI_BAR);
		probe_ent->port[1].scr_addr =
			pci_resource_start(pdev, SIS_SCR_PCI_BAR) + port_map->port2_start;
	}

	pci_set_master(pdev);
	pci_intx(pdev, 1);

	/* FIXME: check ata_device_add return value */
	ata_device_add(probe_ent);
	kfree(probe_ent);

	return 0;

err_out_regions:
	pci_release_regions(pdev);

err_out:
	if (!pci_dev_busy)
		pci_disable_device(pdev);
err_mem:
	kfree(port_map);

	return rc;

}
static void sis_host_stop (struct ata_host_set *host_set)
{
	if (host_set->private_data)
		kfree(host_set->private_data);
	ata_host_stop(host_set);
}

static int __init sis_init(void)
{
	return pci_module_init(&sis_pci_driver);
}

static void __exit sis_exit(void)
{
	pci_unregister_driver(&sis_pci_driver);
}

module_init(sis_init);
module_exit(sis_exit);

