/*
 * Connection tracking state replication for Netfilter
 *
 * (C) 2002-2003, KOVACS Krisztian <hidden@sch.bme.hu>
 * (C) 2003-2004, Harald Welte <laforge@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#ifndef _CT_SYNC_H
#define _CT_SYNC_H

#include <linux/list.h>
#include <linux/types.h>
#include <linux/if_ether.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>

#define CTSYNC_DEBUG 0

/* debugging macros */
#if (CTSYNC_DEBUG > 0)
#define CT_SYNC_DEBUG(x, args...) printk (KERN_DEBUG "ct_sync: %s " x, __FUNCTION__ , ## args)
#define CT_SYNC_ASSERT(x) do {						\
	if (!(x)) 							\
	printk (KERN_ERR "ct_sync: %s:%i(%s) assertion failed.\n", 	\
		__FILE__, __LINE__, __FUNCTION__);			\
} while (0)
#else
#define CT_SYNC_DEBUG(x, args...)
#define CT_SYNC_ASSERT(x, args...)
#endif

#if (CTSYNC_DEBUG >= 2)
#define CT_SYNC_ENTER()		CT_SYNC_DEBUG("enter\n") 
#define CT_SYNC_LEAVE()		CT_SYNC_DEBUG("leave\n")
#define CT_SYNC_DEBUG2(x, args...) printk (KERN_DEBUG "ct_sync: %s " x, __FUNCTION__ , ## args)
#else
#define CT_SYNC_ENTER()
#define CT_SYNC_LEAVE()
#define CT_SYNC_DEBUG2(x, args...)
#endif

#define CT_SYNC_ERR(x, args...) printk(KERN_ERR "ct_sync: %s " x, __FUNCTION__ , ## args)
#define CT_SYNC_INFO(x, args...) printk(KERN_INFO "ct_sync: %s " x, __FUNCTION__ , ## args)

/* predefined sizes */
#define CT_SYNC_CTHELPERSIZE	16	/* maximal length of ct helper name */
#define CT_SYNC_NATHELPERSIZE	16	/* maximal length of NAT helper name */
#define CT_SYNC_PAYLOADSIZE	ETH_DATA_LEN - 20 - 8 - sizeof(struct ct_sync_pkthdr)	/* maximal useful size of sync packets*/

#define CTMSG_SIZEOF(x)		(sizeof(x)+sizeof(struct ct_sync_msghdr))

/* packet header */
struct ct_sync_pkthdr
{
	__u8		version;		/* version number */
	__u8		pkttype;		/* type of packet: sync/NACK */
	__u8		__reserved1;		/* id of the sender node */
	__u8		count;			/* number of msgs in packet */
	__u16		pktseq;			/* packet sequence number */
	__u16		minseq;			/* min seqno master has */
	__u8		flags;			/* flags: 1 == recover */
	__u8		__reserved2;		/* placeholder :) */
	__u16		__reserved3;		/* placeholder :) */
};

/* packet types */
enum ct_sync_pkttype_t
{
	CT_SYNC_PKT_UNKNOWN 		= 0x00,	/* not defined, illegal */
	CT_SYNC_PKT_MASTER_ANNOUNCE	= 0x01, /* master announcement (carrying seqno) */
	CT_SYNC_PKT_SLAVE_INITSYNC	= 0x02, /* slave requests a full re-sync */
	CT_SYNC_PKT_SYNC		= 0x03,	/* normal sync packet */
	CT_SYNC_PKT_NACK		= 0x04,	/* negative ack packet */
	CT_SYNC_PKT_INITSYNC_FOLLOWS	= 0x05,	/* a full re-sync follows */
};

enum ct_sync_pkt_flag_t
{
	CT_SYNC_PKT_F_RECOVER 		= 0x01, /* recovery packet */
};

/* CT_SYNC_PKT_MASTER_ANNOUNCE payload */
struct ct_sync_master_announce
{
	__u32	upsince;			/* master up since */
	__u8	flags;				/* flags */
};

/* CT_SYNC_PKT_SLAVE_INITSYNC payload */
struct ct_sync_slave_initsync
{
	__u32	master;
};

/* CT_SYNC_PKT_NACK doesn't have any payload */

/* definitions for CT_SYNC_PKT_SYNC */
struct ct_sync_pkt
{
	struct ct_sync_pkthdr	hdr;
	__u8			data[CT_SYNC_PAYLOADSIZE];
};

/* CT_SYNC_PKT_SYNC: message header */
struct ct_sync_msghdr
{
	__u8		resource;	/* which table? */
	__u8		type;		/* type of message: ct_sync_msgtype_t */
	__u16		len;		/* length of message */
	__u8		flags;		/* sync flags */
	__u8		res1;
	__u16		res2;
};

/* CT_SYNC_PKT_SYNC: message types */
enum ct_sync_msgtype_t
{
	CT_SYNC_MSG_UPDATE	= 0x01,	/* update conntrack entry 
					   (or create if not exists) */
	CT_SYNC_MSG_DELETE	= 0x02,	/* delete conntrack entry */
};

/* CT_SYNC_PKT_SYNC: resources */
enum ct_sync_resource_t
{
	CT_SYNC_RES_CONNTRACK	= 0x01,
	CT_SYNC_RES_EXPECT	= 0x02,
};

enum ct_sync_upd_flag_t
{
	CTS_UPD_F_NEW 		= 0x01,	/* creates new conntrack */
	CTS_UPD_F_INITSYNC	= 0x02,	/* part of initial sync */
	CTS_UPD_F_INITSYNC_DONE	= 0x04, /* sent after an initsync session is complete */
};

/* linearized conntrack data */
struct ct_sync_conntrack
{
	struct ip_conntrack_tuple	orig, reply;	/* tuples */
	__u32				status;		/* status */
	__u32				expires;	/* timeout */

	/* conntrack helper */
	char				helper[CT_SYNC_CTHELPERSIZE];	/* conntrack helper name */
	union ip_conntrack_help		help;			/* conntrack helper specific info */

	unsigned int			expecting;	/* number of siblings */

	struct ip_conntrack_tuple	master_ct_tuple; /* master ct tuple */

	/* protocol */
	union ip_conntrack_proto	proto;

	/* NAT */
	#ifdef CONFIG_IP_NF_NAT_NEEDED
	int				nat_initialized;
	__u8				nat_num_manips;
	char				nat_helper[CT_SYNC_NATHELPERSIZE];
	union ip_conntrack_nat_help	nat_help;
	struct ip_nat_seq		nat_seq[IP_CT_DIR_MAX];
	#if defined(CONFIG_IP_NF_TARGET_MASQUERADE) || defined(CONFIG_IP_NF_TARGET_MASQUERADE_MODULE)
	char				nat_masq_iface[IFNAMSIZ];
	#endif /* MASQUERADE */
	#endif /* NAT */

	/* mark */
	#ifdef CONFIG_IP_NF_CONNTRACK_MARK
	__u32 				mark;
	#endif
};

/* linearized expectation data */
struct ct_sync_expect
{
	struct ip_conntrack_tuple	expectant;	/* tuple of master conntrack */
	struct ip_conntrack_tuple	sibling;	/* tuple of sibling conntrack, if already arrived */

	struct ip_conntrack_tuple	tuple, mask;	/* expectation tuple and mask */
	__u32				seq;		/* sequence number */
	union ip_conntrack_expect_proto	proto;		/* protocol specific info */
};

#ifdef __KERNEL__

/* statistics */
struct ct_sync_st
{
	unsigned long pkts_tot;		/* total packets */
	unsigned long pkts_ooseq;	/* rx only: total out-of-sequence */
	unsigned long pkts_recover;	/* total recover pkts */

	unsigned long msgs;		/* total messages */
	unsigned long error;		/* total invalid msgs */

	unsigned long upd_tot;		/* total update messages */
	unsigned long upd_new; 		/* total new updates */
	unsigned long upd_initial;	/* total initsync updates */

	unsigned long upd_newbutexist;	/* rx only: new but already exists  */
	unsigned long upd_nothere;	/* rx only: not new but not here */

	unsigned long del_tot;		/* total deletes */
	unsigned long del_nothere;	/* rx only: delete but not here */

	unsigned long exp_upd_tot;	/* total expect updates */
	unsigned long exp_del_tot;	/* total delete updates */
};


struct ct_sync_stat
{
	struct ct_sync_st rx;
	struct ct_sync_st tx;

	unsigned long slave2master;	/* slave->master transitions */
	unsigned long master2slave;	/* master->slave transitions */
	unsigned long master2slave_auto;

	unsigned long initsyncs;	/* master: started initsyncs */
};

DECLARE_PER_CPU(struct ct_sync_stat, ct_sync_stats);

#define CTS_STAT_INC(field)	(__get_cpu_var(ct_sync_stats).field++)

#endif /* _KERNEL_ */

#endif
