#ifndef _CTS_BUFF_H
#define _CTS_BUFF_H

#include <linux/skbuff.h>
#include <linux/list.h>
#include <linux/netfilter_ipv4/ct_sync.h>


#if (CTSYNC_DEBUG >= 4)
#define DUMP_CTS_BUFF(x)						\
	printk(KERN_DEBUG "%p use=%u size=%u len=%u parsed=%p\n", 	\
		x, csb_used(x), csb_size(x), csb_len(x), x->parsed);
#else 
#define DUMP_CTS_BUFF(x)
#endif

struct csb_ring;		/* forward declaration */
struct cts_buff {
	struct list_head list;
	struct csb_ring *ring;	/* ring on which we are */
#if (CTSYNC_DEBUG >= 4)
	unsigned int ringnum;	/* number inside ring */
#endif

	atomic_t use;		/* is somebody still busy writing data? */
	unsigned long firstuse;	/* time in jiffies the first message was written */
	char *tail;		/* first byte after used packet data */
	char *end;		/* last byte of allocated buffer */
	char *parsed; 		/* receiver: packet parsed up to here */
	struct {
		struct ct_sync_pkthdr hdr;
		char data[0];
	} pkt;
};

/* our ring of cts_buffs.  This ring is basically split in four sections:
 * (sent->next, cur->prev): 	cts_buffs that are pending to be sent
 * (cur, alloc):		cts_buffs that are allocated but busy
 * (alloc->next,backlog->prev):	cts_buffs that are free to be overwritten
 * (backlog, sent):		cts_buffs that are saved for retransmissions
 *
 * New packets are allocated by advancing the 'alloc' pointer.
 * Once we finished to complete a new packet, we advance 'cur' and wake up
 * the sender thread.   If we hit 'backlog', we need to allocate a new buffer
 * and put it in the ring.
 *
 * The sender thread then sends a packet and advances 'sent' as well as
 * 'backlog' (to keep a fixed-length backlog). 
 *
 * 'lock' protects the three pointers themselves, as well as the cts_buff
 * pointed to by 'cur'
 */
struct csb_ring {
	struct list_head *alloc;	/* up to here we have already alloced*/
	struct list_head *cur;		/* current csb, not yet fully filled */
	struct list_head *sent;		/* last csb that was already sent */
	struct list_head *backlog;	/* last csb that we have to save */
	__u16 seqno;			/* sequence number */
	spinlock_t lock;
};

/***********************************************************************
 * CTS_BUFF / CSB_RING util functions
 ***********************************************************************/

#if (CTSYNC_DEBUG >= 4)
static inline void
csr_print(struct csb_ring *csr)
{
	int i = 0;
	struct cts_buff *csb;

	list_for_each_entry(csb, csr->cur, list) {
		i++;
	}

	printk(KERN_DEBUG "%s(%p): alloc=%d, cur=%d, sent=%d, backlog=%d (total=%d)\n", __FUNCTION__, csr, 
		((struct cts_buff *)(csr->alloc))->ringnum, 
		((struct cts_buff *)(csr->cur))->ringnum, 
		((struct cts_buff *)(csr->sent))->ringnum, 
		((struct cts_buff *)(csr->backlog))->ringnum, i);
}
#else
#define csr_print(x)
#endif


static inline unsigned char *csb_put(struct cts_buff *csb, unsigned int len)
{
	unsigned char *tmp = csb->tail;
	if (unlikely(csb->tail + len > csb->end))
		return NULL;

	csb->tail += len;
	return tmp;
}

/* return size of buffer in memory*/
static inline unsigned int csb_size(const struct cts_buff *csb)
{
	return (csb->end - (char *) &csb->pkt);
}

/* return size of used bytes within buffer */
static inline unsigned int csb_len(const struct cts_buff *csb)
{
	return (csb->tail - (char *) &csb->pkt);
}

static inline void csb_use_inc(struct cts_buff *csb)
{
	atomic_inc(&csb->use);
}
static inline int csb_used(const struct cts_buff *csb)
{
	return atomic_read(&csb->use);
}

#endif /* _CTS_BUFF_H */
